@@ -14,12 +14,19 @@
 
 import abc, contextlib, enum, os.path, re, tempfile, shlex
 import subprocess
+from typing import List, Tuple
 
 from ..linkers import StaticLinker
 from .. import coredata
 from .. import mlog
 from .. import mesonlib
-from ..mesonlib import EnvironmentException, MesonException, OrderedSet, version_compare, Popen_safe
+from ..mesonlib import (
+    EnvironmentException, MesonException, OrderedSet,
+    version_compare, Popen_safe
+)
+from ..envconfig import (
+    Properties,
+)
 
 """This file contains the data files of all compilers Meson knows
 about. To support a new compiler, add its information below.
@@ -33,6 +40,7 @@ lib_suffixes = ('a', 'lib', 'dll', 'dylib', 'so')
 lang_suffixes = {
     'c': ('c',),
     'cpp': ('cpp', 'cc', 'cxx', 'c++', 'hh', 'hpp', 'ipp', 'hxx'),
+    'cuda': ('cu',),
     # f90, f95, f03, f08 are for free-form fortran ('f90' recommended)
     # f, for, ftn, fpp are for fixed-form fortran ('f' or 'for' recommended)
     'fortran': ('f90', 'f95', 'f03', 'f08', 'f', 'for', 'ftn', 'fpp'),
@@ -54,7 +62,7 @@ clib_langs = ('objcpp', 'cpp', 'objc', 'c', 'fortran',)
 # List of languages that can be linked with C code directly by the linker
 # used in build.py:process_compilers() and build.py:get_dynamic_linker()
 # XXX: Add Rust to this?
-clink_langs = ('d',) + clib_langs
+clink_langs = ('d', 'cuda') + clib_langs
 clink_suffixes = ()
 for _l in clink_langs + ('vala',):
     clink_suffixes += lang_suffixes[_l]
@@ -65,6 +73,7 @@ soregex = re.compile(r'.*\.so(\.[0-9]+)?(\.[0-9]+)?(\.[0-9]+)?$')
 # Environment variables that each lang uses.
 cflags_mapping = {'c': 'CFLAGS',
                   'cpp': 'CXXFLAGS',
+                  'cuda': 'CUFLAGS',
                   'objc': 'OBJCFLAGS',
                   'objcpp': 'OBJCXXFLAGS',
                   'fortran': 'FFLAGS',
@@ -139,6 +148,13 @@ armclang_buildtype_args = {'plain': [],
                            'custom': [],
                            }
 
+cuda_buildtype_args = {'plain': [],
+                       'debug': [],
+                       'debugoptimized': [],
+                       'release': [],
+                       'minsize': [],
+                       }
+
 arm_buildtype_args = {'plain': [],
                       'debug': ['-O0', '--debug'],
                       'debugoptimized': ['-O1', '--debug'],
@@ -147,6 +163,14 @@ arm_buildtype_args = {'plain': [],
                       'custom': [],
                       }
 
+ccrx_buildtype_args = {'plain': [],
+                       'debug': [],
+                       'debugoptimized': [],
+                       'release': [],
+                       'minsize': [],
+                       'custom': [],
+                       }
+
 msvc_buildtype_args = {'plain': [],
                        'debug': ["/ZI", "/Ob0", "/Od", "/RTC1"],
                        'debugoptimized': ["/Zi", "/Ob1"],
@@ -155,6 +179,13 @@ msvc_buildtype_args = {'plain': [],
                        'custom': [],
                        }
 
+pgi_buildtype_args = {'plain': [],
+                      'debug': [],
+                      'debugoptimized': [],
+                      'release': [],
+                      'minsize': [],
+                      'custom': [],
+                      }
 apple_buildtype_linker_args = {'plain': [],
                                'debug': [],
                                'debugoptimized': [],
@@ -179,6 +210,21 @@ arm_buildtype_linker_args = {'plain': [],
                              'custom': [],
                              }
 
+ccrx_buildtype_linker_args = {'plain': [],
+                              'debug': [],
+                              'debugoptimized': [],
+                              'release': [],
+                              'minsize': [],
+                              'custom': [],
+                              }
+pgi_buildtype_linker_args = {'plain': [],
+                             'debug': [],
+                             'debugoptimized': [],
+                             'release': [],
+                             'minsize': [],
+                             'custom': [],
+                             }
+
 msvc_buildtype_linker_args = {'plain': [],
                               'debug': [],
                               'debugoptimized': [],
@@ -295,20 +341,42 @@ gnu_optimization_args = {'0': [],
                          's': ['-Os'],
                          }
 
+ccrx_optimization_args = {'0': ['-optimize=0'],
+                          'g': ['-optimize=0'],
+                          '1': ['-optimize=1'],
+                          '2': ['-optimize=2'],
+                          '3': ['-optimize=max'],
+                          's': ['-optimize=2', '-size']
+                          }
+
 msvc_optimization_args = {'0': [],
                           'g': ['/O0'],
                           '1': ['/O1'],
                           '2': ['/O2'],
-                          '3': ['/O3'],
+                          '3': ['/O2'],
                           's': ['/O1'], # Implies /Os.
                           }
 
+cuda_optimization_args = {'0': [],
+                          'g': ['-O0'],
+                          '1': ['-O1'],
+                          '2': ['-O2'],
+                          '3': ['-O3', '-Otime'],
+                          's': ['-O3', '-Ospace']
+                          }
+
+cuda_debug_args = {False: [],
+                   True: ['-g']}
+
 clike_debug_args = {False: [],
                     True: ['-g']}
 
 msvc_debug_args = {False: [],
                    True: []} # Fixme!
 
+ccrx_debug_args = {False: [],
+                   True: ['-debug']}
+
 base_options = {'b_pch': coredata.UserBooleanOption('b_pch', 'Use precompiled headers', True),
                 'b_lto': coredata.UserBooleanOption('b_lto', 'Use link time optimization', False),
                 'b_sanitize': coredata.UserComboOption('b_sanitize',
@@ -331,6 +399,9 @@ base_options = {'b_pch': coredata.UserBooleanOption('b_pch', 'Use precompiled he
                 'b_staticpic': coredata.UserBooleanOption('b_staticpic',
                                                           'Build static libraries as position independent',
                                                           True),
+                'b_pie': coredata.UserBooleanOption('b_pie',
+                                                    'Build executables as position independent',
+                                                    False),
                 'b_bitcode': coredata.UserBooleanOption('b_bitcode',
                                                         'Generate and embed bitcode (only macOS and iOS)',
                                                         False),
@@ -424,9 +495,9 @@ def get_base_compile_args(options, compiler):
     try:
         pgo_val = options['b_pgo'].value
         if pgo_val == 'generate':
-            args.append('-fprofile-generate')
+            args.extend(compiler.get_profile_generate_args())
         elif pgo_val == 'use':
-            args.append('-fprofile-use')
+            args.extend(compiler.get_profile_use_args())
     except KeyError:
         pass
     try:
@@ -470,9 +541,9 @@ def get_base_link_args(options, linker, is_shared_module):
     try:
         pgo_val = options['b_pgo'].value
         if pgo_val == 'generate':
-            args.append('-fprofile-generate')
+            args.extend(linker.get_profile_generate_args())
         elif pgo_val == 'use':
-            args.append('-fprofile-use')
+            args.extend(linker.get_profile_use_args())
     except KeyError:
         pass
     try:
@@ -503,6 +574,31 @@ def get_base_link_args(options, linker, is_shared_module):
         pass
     return args
 
+def prepare_rpaths(raw_rpaths, build_dir, from_dir):
+    internal_format_rpaths = [evaluate_rpath(p, build_dir, from_dir) for p in raw_rpaths]
+    ordered_rpaths = order_rpaths(internal_format_rpaths)
+    return ordered_rpaths
+
+def order_rpaths(rpath_list):
+    # We want rpaths that point inside our build dir to always override
+    # those pointing to other places in the file system. This is so built
+    # binaries prefer our libraries to the ones that may lie somewhere
+    # in the file system, such as /lib/x86_64-linux-gnu.
+    #
+    # The correct thing to do here would be C++'s std::stable_partition.
+    # Python standard library does not have it, so replicate it with
+    # sort, which is guaranteed to be stable.
+    return sorted(rpath_list, key=os.path.isabs)
+
+def evaluate_rpath(p, build_dir, from_dir):
+    if p == from_dir:
+        return '' # relpath errors out in this case
+    elif os.path.isabs(p):
+        return p # These can be outside of build dir.
+    else:
+        return os.path.relpath(os.path.join(build_dir, p), os.path.join(build_dir, from_dir))
+
+
 class CrossNoRunException(MesonException):
     pass
 
@@ -554,7 +650,10 @@ class CompilerArgs(list):
     dedup2_suffixes = ()
     dedup2_args = ()
     # Arg prefixes and args that must be de-duped by returning 1
-    dedup1_prefixes = ('-l', '-Wl,-l')
+    #
+    # NOTE: not thorough. A list of potential corner cases can be found in
+    # https://github.com/mesonbuild/meson/pull/4593#pullrequestreview-182016038
+    dedup1_prefixes = ('-l', '-Wl,-l', '-Wl,--export-dynamic')
     dedup1_suffixes = ('.lib', '.dll', '.so', '.dylib', '.a')
     # Match a .so of the form path/to/libfoo.so.0.1.0
     # Only UNIX shared libraries require this. Others have a fixed extension.
@@ -616,7 +715,6 @@ class CompilerArgs(list):
         to recursively search for symbols in the libraries. This is not needed
         with other linkers.
         '''
-
         # A standalone argument must never be deduplicated because it is
         # defined by what comes _after_ it. Thus dedupping this:
         # -D FOO -D BAR
@@ -691,6 +789,17 @@ class CompilerArgs(list):
         for elem in iterable:
             self.append_direct(elem)
 
+    def extend_preserving_lflags(self, iterable):
+        normal_flags = []
+        lflags = []
+        for i in iterable:
+            if i.startswith('-l') or i.startswith('-L'):
+                lflags.append(i)
+            else:
+                normal_flags.append(i)
+        self.extend(normal_flags)
+        self.extend_direct(lflags)
+
     def __add__(self, args):
         new = CompilerArgs(self, self.compiler)
         new += args
@@ -813,10 +922,13 @@ class Compiler:
     def compute_int(self, expression, low, high, guess, prefix, env, extra_args, dependencies):
         raise EnvironmentException('%s does not support compute_int ' % self.get_id())
 
-    def has_members(self, typename, membernames, prefix, env, extra_args=None, dependencies=None):
+    def compute_parameters_with_absolute_paths(self, parameter_list, build_dir):
+        raise EnvironmentException('%s does not support compute_parameters_with_absolute_paths ' % self.get_id())
+
+    def has_members(self, typename, membernames, prefix, env, *, extra_args=None, dependencies=None):
         raise EnvironmentException('%s does not support has_member(s) ' % self.get_id())
 
-    def has_type(self, typename, prefix, env, extra_args, dependencies=None):
+    def has_type(self, typename, prefix, env, extra_args, *, dependencies=None):
         raise EnvironmentException('%s does not support has_type ' % self.get_id())
 
     def symbols_have_underscore_prefix(self, env):
@@ -843,6 +955,9 @@ class Compiler:
     def get_linker_always_args(self):
         return []
 
+    def get_linker_lib_prefix(self):
+        return ''
+
     def gen_import_library_args(self, implibname):
         """
         Used only on Windows for libraries that need an import library.
@@ -897,23 +1012,44 @@ class Compiler:
 
     def get_options(self):
         opts = {} # build afresh every time
-
-        # Take default values from env variables.
-        compile_args, link_args = self.get_args_from_envvars()
         description = 'Extra arguments passed to the {}'.format(self.get_display_language())
         opts.update({
             self.language + '_args': coredata.UserArrayOption(
                 self.language + '_args',
                 description + ' compiler',
-                compile_args, shlex_split=True, user_input=True, allow_dups=True),
+                [], shlex_split=True, user_input=True, allow_dups=True),
             self.language + '_link_args': coredata.UserArrayOption(
                 self.language + '_link_args',
                 description + ' linker',
-                link_args, shlex_split=True, user_input=True, allow_dups=True),
+                [], shlex_split=True, user_input=True, allow_dups=True),
         })
 
         return opts
 
+    def get_and_default_options(self, properties: Properties):
+        """
+        Take default values from env variables and/or config files.
+        """
+        opts = self.get_options()
+
+        if properties.fallback:
+            # Get from env vars.
+            compile_args, link_args = self.get_args_from_envvars()
+        else:
+            compile_args = []
+            link_args = []
+
+        for k, o in opts.items():
+            if k in properties:
+                # Get from configuration files.
+                o.set_value(properties[k])
+            elif k == self.language + '_args':
+                o.set_value(compile_args)
+            elif k == self.language + '_link_args':
+                o.set_value(link_args)
+
+        return opts
+
     def get_option_compile_args(self, options):
         return []
 
@@ -968,18 +1104,6 @@ class Compiler:
             'Language {} does not support has_multi_link_arguments.'.format(
                 self.get_display_language()))
 
-    def get_cross_extra_flags(self, environment, link):
-        extra_flags = []
-        if self.is_cross and environment:
-            if 'properties' in environment.cross_info.config:
-                props = environment.cross_info.config['properties']
-                lang_args_key = self.language + '_args'
-                extra_flags += props.get(lang_args_key, [])
-                lang_link_args_key = self.language + '_link_args'
-                if link:
-                    extra_flags += props.get(lang_link_args_key, [])
-        return extra_flags
-
     def _get_compile_output(self, dirname, mode):
         # In pre-processor mode, the output is sent to stdout and discarded
         if mode == 'preprocess':
@@ -1088,46 +1212,30 @@ class Compiler:
     def get_instruction_set_args(self, instruction_set):
         return None
 
-    def build_osx_rpath_args(self, build_dir, rpath_paths, build_rpath):
-        # Ensure that there is enough space for large RPATHs and install_name
-        args = ['-Wl,-headerpad_max_install_names']
-        if not rpath_paths and not build_rpath:
-            return args
-        # On OSX, rpaths must be absolute.
-        abs_rpaths = [os.path.join(build_dir, p) for p in rpath_paths]
-        if build_rpath != '':
-            abs_rpaths.append(build_rpath)
-        # Need to deduplicate abs_rpaths, as rpath_paths and
-        # build_rpath are not guaranteed to be disjoint sets
-        args += ['-Wl,-rpath,' + rp for rp in OrderedSet(abs_rpaths)]
-        return args
-
     def build_unix_rpath_args(self, build_dir, from_dir, rpath_paths, build_rpath, install_rpath):
         if not rpath_paths and not install_rpath and not build_rpath:
             return []
-        # The rpaths we write must be relative, because otherwise
-        # they have different length depending on the build
+        args = []
+        if mesonlib.is_osx():
+            # Ensure that there is enough space for install_name_tool in-place editing of large RPATHs
+            args.append('-Wl,-headerpad_max_install_names')
+            # @loader_path is the equivalent of $ORIGIN on macOS
+            # https://stackoverflow.com/q/26280738
+            origin_placeholder = '@loader_path'
+        else:
+            origin_placeholder = '$ORIGIN'
+        # The rpaths we write must be relative if they point to the build dir,
+        # because otherwise they have different length depending on the build
         # directory. This breaks reproducible builds.
-        rel_rpaths = []
-        for p in rpath_paths:
-            if p == from_dir:
-                relative = '' # relpath errors out in this case
-            else:
-                relative = os.path.relpath(os.path.join(build_dir, p), os.path.join(build_dir, from_dir))
-            rel_rpaths.append(relative)
-        paths = ':'.join([os.path.join('$ORIGIN', p) for p in rel_rpaths])
+        processed_rpaths = prepare_rpaths(rpath_paths, build_dir, from_dir)
+        # Need to deduplicate rpaths, as macOS's install_name_tool
+        # is *very* allergic to duplicate -delete_rpath arguments
+        # when calling depfixer on installation.
+        all_paths = OrderedSet([os.path.join(origin_placeholder, p) for p in processed_rpaths])
         # Build_rpath is used as-is (it is usually absolute).
         if build_rpath != '':
-            if paths != '':
-                paths += ':'
-            paths += build_rpath
-        if len(paths) < len(install_rpath):
-            padding = 'X' * (len(install_rpath) - len(paths))
-            if not paths:
-                paths = padding
-            else:
-                paths = paths + ':' + padding
-        args = []
+            all_paths.add(build_rpath)
+
         if mesonlib.is_dragonflybsd() or mesonlib.is_openbsd():
             # This argument instructs the compiler to record the value of
             # ORIGIN in the .dynamic section of the elf. On Linux this is done
@@ -1135,7 +1243,23 @@ class Compiler:
             # $ORIGIN in the runtime path will be undefined and any binaries
             # linked against local libraries will fail to resolve them.
             args.append('-Wl,-z,origin')
-        args.append('-Wl,-rpath,' + paths)
+
+        if mesonlib.is_osx():
+            # macOS does not support colon-separated strings in LC_RPATH,
+            # hence we have to pass each path component individually
+            args += ['-Wl,-rpath,' + rp for rp in all_paths]
+        else:
+            # In order to avoid relinking for RPATH removal, the binary needs to contain just
+            # enough space in the ELF header to hold the final installation RPATH.
+            paths = ':'.join(all_paths)
+            if len(paths) < len(install_rpath):
+                padding = 'X' * (len(install_rpath) - len(paths))
+                if not paths:
+                    paths = padding
+                else:
+                    paths = paths + ':' + padding
+            args.append('-Wl,-rpath,' + paths)
+
         if get_compiler_is_linuxlike(self):
             # Rpaths to use while linking must be absolute. These are not
             # written to the binary. Needed only with GNU ld:
@@ -1174,6 +1298,46 @@ class Compiler:
         raise EnvironmentException(
             'Language {} does not support function attributes.'.format(self.get_display_language()))
 
+    def get_pic_args(self):
+        m = 'Language {} does not support position-independent code'
+        raise EnvironmentException(m.format(self.get_display_language()))
+
+    def get_pie_args(self):
+        m = 'Language {} does not support position-independent executable'
+        raise EnvironmentException(m.format(self.get_display_language()))
+
+    def get_pie_link_args(self):
+        m = 'Language {} does not support position-independent executable'
+        raise EnvironmentException(m.format(self.get_display_language()))
+
+    def get_argument_syntax(self):
+        """Returns the argument family type.
+
+        Compilers fall into families if they try to emulate the command line
+        interface of another compiler. For example, clang is in the GCC family
+        since it accepts most of the same arguments as GCC. ICL (ICC on
+        windows) is in the MSVC family since it accepts most of the same
+        arguments as MSVC.
+        """
+        return 'other'
+
+    def get_profile_generate_args(self):
+        raise EnvironmentException(
+            '%s does not support get_profile_generate_args ' % self.get_id())
+
+    def get_profile_use_args(self):
+        raise EnvironmentException(
+            '%s does not support get_profile_use_args ' % self.get_id())
+
+    def get_undefined_link_args(self):
+        '''
+        Get args for allowing undefined symbols when linking to a shared library
+        '''
+        return []
+
+    def remove_linkerlike_args(self, args):
+        return [x for x in args if not x.startswith('-Wl')]
+
 
 @enum.unique
 class CompilerType(enum.Enum):
@@ -1191,6 +1355,12 @@ class CompilerType(enum.Enum):
     ICC_OSX = 21
     ICC_WIN = 22
 
+    ARM_WIN = 30
+
+    CCRX_WIN = 40
+
+    PGI_STANDARD = 50
+
     @property
     def is_standard_compiler(self):
         return self.name in ('GCC_STANDARD', 'CLANG_STANDARD', 'ICC_STANDARD')
@@ -1201,7 +1371,7 @@ class CompilerType(enum.Enum):
 
     @property
     def is_windows_compiler(self):
-        return self.name in ('GCC_MINGW', 'GCC_CYGWIN', 'CLANG_MINGW', 'ICC_WIN')
+        return self.name in ('GCC_MINGW', 'GCC_CYGWIN', 'CLANG_MINGW', 'ICC_WIN', 'ARM_WIN', 'CCRX_WIN')
 
 
 def get_macos_dylib_install_name(prefix, shlib_name, suffix, soversion):
@@ -1314,9 +1484,13 @@ class GnuLikeCompiler(abc.ABC):
     def __init__(self, compiler_type):
         self.compiler_type = compiler_type
         self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage',
-                             'b_ndebug', 'b_staticpic', 'b_asneeded']
-        if not self.compiler_type.is_osx_compiler:
+                             'b_ndebug', 'b_staticpic', 'b_pie']
+        if (not self.compiler_type.is_osx_compiler and
+                not self.compiler_type.is_windows_compiler and
+                not mesonlib.is_openbsd()):
             self.base_options.append('b_lundef')
+        if not self.compiler_type.is_windows_compiler:
+            self.base_options.append('b_asneeded')
         # All GCC-like backends can do assembly
         self.can_compile_suffixes.add('s')
 
@@ -1335,6 +1509,12 @@ class GnuLikeCompiler(abc.ABC):
             return [] # On Window and OS X, pic is always on.
         return ['-fPIC']
 
+    def get_pie_args(self):
+        return ['-fPIE']
+
+    def get_pie_link_args(self):
+        return ['-pie']
+
     def get_buildtype_args(self, buildtype):
         return gnulike_buildtype_args[buildtype]
 
@@ -1389,24 +1569,65 @@ class GnuLikeCompiler(abc.ABC):
     def gnu_symbol_visibility_args(self, vistype):
         return gnu_symbol_visibility_args[vistype]
 
+    def gen_vs_module_defs_args(self, defsfile):
+        if not isinstance(defsfile, str):
+            raise RuntimeError('Module definitions file should be str')
+        # On Windows targets, .def files may be specified on the linker command
+        # line like an object file.
+        if self.compiler_type.is_windows_compiler:
+            return [defsfile]
+        # For other targets, discard the .def file.
+        return []
+
+    def get_argument_syntax(self):
+        return 'gcc'
+
+    def get_profile_generate_args(self):
+        return ['-fprofile-generate']
+
+    def get_profile_use_args(self):
+        return ['-fprofile-use', '-fprofile-correction']
+
+    def get_allow_undefined_link_args(self):
+        if self.compiler_type.is_osx_compiler:
+            # Apple ld
+            return ['-Wl,-undefined,dynamic_lookup']
+        elif self.compiler_type.is_windows_compiler:
+            # For PE/COFF this is impossible
+            return []
+        else:
+            # GNU ld and LLVM lld
+            return ['-Wl,--allow-shlib-undefined']
+
+    def get_gui_app_args(self, value):
+        if self.compiler_type.is_windows_compiler:
+            return ['-mwindows' if value else '-mconsole']
+        return []
+
+    def compute_parameters_with_absolute_paths(self, parameter_list, build_dir):
+        for idx, i in enumerate(parameter_list):
+            if i[:2] == '-I' or i[:2] == '-L':
+                parameter_list[idx] = i[:2] + os.path.normpath(os.path.join(build_dir, i[2:]))
+
+        return parameter_list
 
 class GnuCompiler(GnuLikeCompiler):
     """
     GnuCompiler represents an actual GCC in its many incarnations.
     Compilers imitating GCC (Clang/Intel) should use the GnuLikeCompiler ABC.
     """
-    def __init__(self, compiler_type, defines):
+    def __init__(self, compiler_type, defines: dict):
         super().__init__(compiler_type)
         self.id = 'gcc'
         self.defines = defines or {}
         self.base_options.append('b_colorout')
 
-    def get_colorout_args(self, colortype):
+    def get_colorout_args(self, colortype: str) -> List[str]:
         if mesonlib.version_compare(self.version, '>=4.9.0'):
             return gnu_color_args[colortype][:]
         return []
 
-    def get_warn_args(self, level):
+    def get_warn_args(self, level: str) -> list:
         args = super().get_warn_args(level)
         if mesonlib.version_compare(self.version, '<4.8.0') and '-Wpedantic' in args:
             # -Wpedantic was added in 4.8.0
@@ -1414,38 +1635,70 @@ class GnuCompiler(GnuLikeCompiler):
             args[args.index('-Wpedantic')] = '-pedantic'
         return args
 
-    def has_builtin_define(self, define):
+    def has_builtin_define(self, define: str) -> bool:
         return define in self.defines
 
     def get_builtin_define(self, define):
         if define in self.defines:
             return self.defines[define]
 
-    def get_optimization_args(self, optimization_level):
+    def get_optimization_args(self, optimization_level: str):
         return gnu_optimization_args[optimization_level]
 
-    def get_pch_suffix(self):
+    def get_pch_suffix(self) -> str:
         return 'gch'
 
-    def gen_vs_module_defs_args(self, defsfile):
-        if not isinstance(defsfile, str):
-            raise RuntimeError('Module definitions file should be str')
-        # On Windows targets, .def files may be specified on the linker command
-        # line like an object file.
-        if self.compiler_type in (CompilerType.GCC_CYGWIN, CompilerType.GCC_MINGW):
-            return [defsfile]
-        # For other targets, discard the .def file.
-        return []
-
-    def get_gui_app_args(self, value):
-        if self.compiler_type in (CompilerType.GCC_CYGWIN, CompilerType.GCC_MINGW) and value:
-            return ['-mwindows']
-        return []
-
-    def openmp_flags(self):
+    def openmp_flags(self) -> List[str]:
         return ['-fopenmp']
 
 
+class PGICompiler:
+    def __init__(self, compiler_type=None):
+        self.id = 'pgi'
+        self.compiler_type = compiler_type
+
+        default_warn_args = ['-Minform=inform']
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
+                          '2': default_warn_args,
+                          '3': default_warn_args}
+
+    def get_module_incdir_args(self) -> Tuple[str]:
+        return ('-module', )
+
+    def get_no_warn_args(self) -> List[str]:
+        return ['-silent']
+
+    def openmp_flags(self) -> List[str]:
+        return ['-mp']
+
+    def get_buildtype_args(self, buildtype: str) -> List[str]:
+        return pgi_buildtype_args[buildtype]
+
+    def get_buildtype_linker_args(self, buildtype: str) -> List[str]:
+        return pgi_buildtype_linker_args[buildtype]
+
+    def get_optimization_args(self, optimization_level: str):
+        return clike_optimization_args[optimization_level]
+
+    def get_debug_args(self, is_debug: bool):
+        return clike_debug_args[is_debug]
+
+    def compute_parameters_with_absolute_paths(self, parameter_list: List[str], build_dir: str):
+        for idx, i in enumerate(parameter_list):
+            if i[:2] == '-I' or i[:2] == '-L':
+                parameter_list[idx] = i[:2] + os.path.normpath(os.path.join(build_dir, i[2:]))
+
+    def get_allow_undefined_link_args(self):
+        return []
+
+    def get_dependency_gen_args(self, outtarget, outfile):
+        return []
+
+    def get_always_args(self):
+        return []
+
+
 class ElbrusCompiler(GnuCompiler):
     # Elbrus compiler is nearly like GCC, but does not support
     # PCH, LTO, sanitizers and color output as of version 1.21.x.
@@ -1518,7 +1771,7 @@ class ClangCompiler(GnuLikeCompiler):
             myargs + args,
             env)
 
-    def has_function(self, funcname, prefix, env, extra_args=None, dependencies=None):
+    def has_function(self, funcname, prefix, env, *, extra_args=None, dependencies=None):
         if extra_args is None:
             extra_args = []
         # Starting with XCode 8, we need to pass this to force linker
@@ -1527,7 +1780,8 @@ class ClangCompiler(GnuLikeCompiler):
         # https://github.com/Homebrew/homebrew-core/issues/3727
         if self.compiler_type.is_osx_compiler and version_compare(self.version, '>=8.0'):
             extra_args.append('-Wl,-no_weak_imports')
-        return super().has_function(funcname, prefix, env, extra_args, dependencies)
+        return super().has_function(funcname, prefix, env, extra_args=extra_args,
+                                    dependencies=dependencies)
 
     def openmp_flags(self):
         if version_compare(self.version, '>=3.8.0'):
@@ -1540,7 +1794,7 @@ class ClangCompiler(GnuLikeCompiler):
 
 
 class ArmclangCompiler:
-    def __init__(self):
+    def __init__(self, compiler_type):
         if not self.is_cross:
             raise EnvironmentException('armclang supports only cross-compilation.')
         # Check whether 'armlink.exe' is available in path
@@ -1559,13 +1813,14 @@ class ArmclangCompiler:
             EnvironmentException('armlink version string not found')
         # Using the regular expression from environment.search_version,
         # which is used for searching compiler version
-        version_regex = '(?<!(\d|\.))(\d{1,2}(\.\d+)+(-[a-zA-Z0-9]+)?)'
+        version_regex = r'(?<!(\d|\.))(\d{1,2}(\.\d+)+(-[a-zA-Z0-9]+)?)'
         linker_ver = re.search(version_regex, ver_str)
         if linker_ver:
             linker_ver = linker_ver.group(0)
         if not version_compare(self.version, '==' + linker_ver):
             raise EnvironmentException('armlink version does not match with compiler version')
         self.id = 'armclang'
+        self.compiler_type = compiler_type
         self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage',
                              'b_ndebug', 'b_staticpic', 'b_colorout']
         # Assembly
@@ -1618,18 +1873,47 @@ class ArmclangCompiler:
     def get_debug_args(self, is_debug):
         return clike_debug_args[is_debug]
 
+    def gen_export_dynamic_link_args(self, env):
+        """
+        The args for export dynamic
+        """
+        return ['--export_dynamic']
 
-# Tested on linux for ICC 14.0.3, 15.0.6, 16.0.4, 17.0.1
+    def gen_import_library_args(self, implibname):
+        """
+        The args of the outputted import library
+
+        ArmLinker's symdefs output can be used as implib
+        """
+        return ['--symdefs=' + implibname]
+
+    def compute_parameters_with_absolute_paths(self, parameter_list, build_dir):
+        for idx, i in enumerate(parameter_list):
+            if i[:2] == '-I' or i[:2] == '-L':
+                parameter_list[idx] = i[:2] + os.path.normpath(os.path.join(build_dir, i[2:]))
+
+        return parameter_list
+
+
+# Tested on linux for ICC 14.0.3, 15.0.6, 16.0.4, 17.0.1, 19.0.0
 class IntelCompiler(GnuLikeCompiler):
+
     def __init__(self, compiler_type):
         super().__init__(compiler_type)
+        # As of 19.0.0 ICC doesn't have sanitizer, color, or lto support.
+        #
+        # It does have IPO, which serves much the same purpose as LOT, but
+        # there is an unfortunate rule for using IPO (you can't control the
+        # name of the output file) which break assumptions meson makes
+        self.base_options = ['b_pch', 'b_lundef', 'b_asneeded', 'b_pgo',
+                             'b_coverage', 'b_ndebug', 'b_staticpic', 'b_pie']
         self.id = 'intel'
         self.lang_header = 'none'
 
     def get_optimization_args(self, optimization_level):
         return clike_optimization_args[optimization_level]
 
-    def get_pch_suffix(self):
+    def get_pch_suffix(self) -> str:
         return 'pchi'
 
     def get_pch_use_args(self, pch_dir, header):
@@ -1639,25 +1923,45 @@ class IntelCompiler(GnuLikeCompiler):
     def get_pch_name(self, header_name):
         return os.path.basename(header_name) + '.' + self.get_pch_suffix()
 
-    def openmp_flags(self):
+    def openmp_flags(self) -> List[str]:
         if version_compare(self.version, '>=15.0.0'):
             return ['-qopenmp']
         else:
             return ['-openmp']
 
-    def has_arguments(self, args, env, code, mode):
-        # -diag-error 10148 is required to catch invalid -W options
-        return super().has_arguments(args + ['-diag-error', '10006', '-diag-error', '10148'], env, code, mode)
+    def compiles(self, *args, **kwargs):
+        # This covers a case that .get('foo', []) doesn't, that extra_args is
+        # defined and is None
+        extra_args = kwargs.get('extra_args') or []
+        kwargs['extra_args'] = [
+            extra_args,
+            '-diag-error', '10006',  # ignoring unknown option
+            '-diag-error', '10148',  # Option not supported
+            '-diag-error', '10155',  # ignoring argument required
+            '-diag-error', '10156',  # ignoring not argument allowed
+            '-diag-error', '10157',  # Ignoring argument of the wrong type
+            '-diag-error', '10158',  # Argument must be separate. Can be hit by trying an option like -foo-bar=foo when -foo=bar is a valid option but -foo-bar isn't
+            '-diag-error', '1292',   # unknown __attribute__
+        ]
+        return super().compiles(*args, **kwargs)
+
+    def get_profile_generate_args(self):
+        return ['-prof-gen=threadsafe']
+
+    def get_profile_use_args(self):
+        return ['-prof-use']
 
 
 class ArmCompiler:
     # Functionality that is common to all ARM family compilers.
-    def __init__(self):
+    def __init__(self, compiler_type):
         if not self.is_cross:
             raise EnvironmentException('armcc supports only cross-compilation.')
         self.id = 'arm'
+        self.compiler_type = compiler_type
         default_warn_args = []
-        self.warn_args = {'1': default_warn_args,
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
                           '2': default_warn_args + [],
                           '3': default_warn_args + []}
         # Assembly
@@ -1724,3 +2028,112 @@ class ArmCompiler:
 
     def get_debug_args(self, is_debug):
         return clike_debug_args[is_debug]
+
+    def compute_parameters_with_absolute_paths(self, parameter_list, build_dir):
+        for idx, i in enumerate(parameter_list):
+            if i[:2] == '-I' or i[:2] == '-L':
+                parameter_list[idx] = i[:2] + os.path.normpath(os.path.join(build_dir, i[2:]))
+
+        return parameter_list
+
+class CcrxCompiler:
+    def __init__(self, compiler_type):
+        if not self.is_cross:
+            raise EnvironmentException('ccrx supports only cross-compilation.')
+        # Check whether 'rlink.exe' is available in path
+        self.linker_exe = 'rlink.exe'
+        args = '--version'
+        try:
+            p, stdo, stderr = Popen_safe(self.linker_exe, args)
+        except OSError as e:
+            err_msg = 'Unknown linker\nRunning "{0}" gave \n"{1}"'.format(' '.join([self.linker_exe] + [args]), e)
+            raise EnvironmentException(err_msg)
+        self.id = 'ccrx'
+        self.compiler_type = compiler_type
+        # Assembly
+        self.can_compile_suffixes.update('s')
+        default_warn_args = []
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
+                          '2': default_warn_args + [],
+                          '3': default_warn_args + []}
+
+    def can_linker_accept_rsp(self):
+        return False
+
+    def get_pic_args(self):
+        # PIC support is not enabled by default for CCRX,
+        # if users want to use it, they need to add the required arguments explicitly
+        return []
+
+    def get_buildtype_args(self, buildtype):
+        return ccrx_buildtype_args[buildtype]
+
+    def get_buildtype_linker_args(self, buildtype):
+        return ccrx_buildtype_linker_args[buildtype]
+
+    # Override CCompiler.get_std_shared_lib_link_args
+    def get_std_shared_lib_link_args(self):
+        return []
+
+    def get_pch_suffix(self):
+        return 'pch'
+
+    def get_pch_use_args(self, pch_dir, header):
+        return []
+
+    # Override CCompiler.get_dependency_gen_args
+    def get_dependency_gen_args(self, outtarget, outfile):
+        return []
+
+    # Override CCompiler.build_rpath_args
+    def build_rpath_args(self, build_dir, from_dir, rpath_paths, build_rpath, install_rpath):
+        return []
+
+    def thread_flags(self, env):
+        return []
+
+    def thread_link_flags(self, env):
+        return []
+
+    def get_linker_exelist(self):
+        return [self.linker_exe]
+
+    def get_linker_lib_prefix(self):
+        return '-lib='
+
+    def get_coverage_args(self):
+        return []
+
+    def get_coverage_link_args(self):
+        return []
+
+    def get_optimization_args(self, optimization_level):
+        return ccrx_optimization_args[optimization_level]
+
+    def get_debug_args(self, is_debug):
+        return ccrx_debug_args[is_debug]
+
+    @classmethod
+    def unix_args_to_native(cls, args):
+        result = []
+        for i in args:
+            if i.startswith('-D'):
+                i = '-define=' + i[2:]
+            if i.startswith('-I'):
+                i = '-include=' + i[2:]
+            if i.startswith('-Wl,-rpath='):
+                continue
+            elif i == '--print-search-dirs':
+                continue
+            elif i.startswith('-L'):
+                continue
+            result.append(i)
+        return result
+
+    def compute_parameters_with_absolute_paths(self, parameter_list, build_dir):
+        for idx, i in enumerate(parameter_list):
+            if i[:9] == '-include=':
+                parameter_list[idx] = i[:9] + os.path.normpath(os.path.join(build_dir, i[9:]))
+
+        return parameter_list
