@@ -19,17 +19,19 @@ from .. import coredata
 from .. import mlog
 from ..mesonlib import MesonException, version_compare
 
-from .c import CCompiler, VisualStudioCCompiler
+from .c import CCompiler, VisualStudioCCompiler, ClangClCCompiler
 from .compilers import (
-    CompilerType,
     gnu_winlibs,
     msvc_winlibs,
+    CompilerType,
     ClangCompiler,
     GnuCompiler,
     ElbrusCompiler,
     IntelCompiler,
+    PGICompiler,
     ArmCompiler,
     ArmclangCompiler,
+    CcrxCompiler,
 )
 from .c_function_attributes import CXX_FUNC_ATTRIBUTES
 
@@ -61,9 +63,11 @@ class CPPCompiler(CCompiler):
         # too strict without this and always fails.
         return super().get_compiler_check_args() + ['-fpermissive']
 
-    def has_header_symbol(self, hname, symbol, prefix, env, extra_args=None, dependencies=None):
+    def has_header_symbol(self, hname, symbol, prefix, env, *, extra_args=None, dependencies=None):
         # Check if it's a C-like symbol
-        if super().has_header_symbol(hname, symbol, prefix, env, extra_args, dependencies):
+        if super().has_header_symbol(hname, symbol, prefix, env,
+                                     extra_args=extra_args,
+                                     dependencies=dependencies):
             return True
         # Check if it's a class or a template
         if extra_args is None:
@@ -73,7 +77,8 @@ class CPPCompiler(CCompiler):
         #include <{header}>
         using {symbol};
         int main () {{ return 0; }}'''
-        return self.compiles(t.format(**fargs), env, extra_args, dependencies)
+        return self.compiles(t.format(**fargs), env, extra_args=extra_args,
+                             dependencies=dependencies)
 
     def _test_cpp_std_arg(self, cpp_std_value):
         # Test whether the compiler understands a -std=XY argument
@@ -130,7 +135,8 @@ class ClangCPPCompiler(ClangCompiler, CPPCompiler):
         CPPCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
         ClangCompiler.__init__(self, compiler_type)
         default_warn_args = ['-Wall', '-Winvalid-pch', '-Wnon-virtual-dtor']
-        self.warn_args = {'1': default_warn_args,
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
                           '3': default_warn_args + ['-Wextra', '-Wpedantic']}
 
@@ -157,11 +163,12 @@ class ClangCPPCompiler(ClangCompiler, CPPCompiler):
 
 
 class ArmclangCPPCompiler(ArmclangCompiler, CPPCompiler):
-    def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwargs):
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrapper=None, **kwargs):
         CPPCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
-        ArmclangCompiler.__init__(self)
+        ArmclangCompiler.__init__(self, compiler_type)
         default_warn_args = ['-Wall', '-Winvalid-pch', '-Wnon-virtual-dtor']
-        self.warn_args = {'1': default_warn_args,
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
                           '3': default_warn_args + ['-Wextra', '-Wpedantic']}
 
@@ -189,7 +196,8 @@ class GnuCPPCompiler(GnuCompiler, CPPCompiler):
         CPPCompiler.__init__(self, exelist, version, is_cross, exe_wrap, **kwargs)
         GnuCompiler.__init__(self, compiler_type, defines)
         default_warn_args = ['-Wall', '-Winvalid-pch', '-Wnon-virtual-dtor']
-        self.warn_args = {'1': default_warn_args,
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
                           '3': default_warn_args + ['-Wextra', '-Wpedantic']}
 
@@ -202,7 +210,7 @@ class GnuCPPCompiler(GnuCompiler, CPPCompiler):
                      'cpp_debugstl': coredata.UserBooleanOption('cpp_debugstl',
                                                                 'STL debug mode',
                                                                 False)})
-        if self.compiler_type == CompilerType.GCC_MINGW:
+        if self.compiler_type.is_windows_compiler:
             opts.update({
                 'cpp_winlibs': coredata.UserArrayOption('cpp_winlibs', 'Standard Win libraries to link against',
                                                         gnu_winlibs), })
@@ -218,7 +226,7 @@ class GnuCPPCompiler(GnuCompiler, CPPCompiler):
         return args
 
     def get_option_link_args(self, options):
-        if self.compiler_type == CompilerType.GCC_MINGW:
+        if self.compiler_type.is_windows_compiler:
             return options['cpp_winlibs'].value[:]
         return []
 
@@ -229,6 +237,12 @@ class GnuCPPCompiler(GnuCompiler, CPPCompiler):
         return ['-lstdc++']
 
 
+class PGICPPCompiler(PGICompiler, CPPCompiler):
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwargs):
+        CPPCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
+        PGICompiler.__init__(self, CompilerType.PGI_STANDARD)
+
+
 class ElbrusCPPCompiler(GnuCPPCompiler, ElbrusCompiler):
     def __init__(self, exelist, version, compiler_type, is_cross, exe_wrapper=None, defines=None, **kwargs):
         GnuCPPCompiler.__init__(self, exelist, version, compiler_type, is_cross, exe_wrapper, defines, **kwargs)
@@ -245,11 +259,13 @@ class ElbrusCPPCompiler(GnuCPPCompiler, ElbrusCompiler):
 
     # Elbrus C++ compiler does not have lchmod, but there is only linker warning, not compiler error.
     # So we should explicitly fail at this case.
-    def has_function(self, funcname, prefix, env, extra_args=None, dependencies=None):
+    def has_function(self, funcname, prefix, env, *, extra_args=None, dependencies=None):
         if funcname == 'lchmod':
             return False
         else:
-            return super().has_function(funcname, prefix, env, extra_args, dependencies)
+            return super().has_function(funcname, prefix, env,
+                                        extra_args=extra_args,
+                                        dependencies=dependencies)
 
 
 class IntelCPPCompiler(IntelCompiler, CPPCompiler):
@@ -259,7 +275,8 @@ class IntelCPPCompiler(IntelCompiler, CPPCompiler):
         self.lang_header = 'c++-header'
         default_warn_args = ['-Wall', '-w3', '-diag-disable:remark',
                              '-Wpch-messages', '-Wnon-virtual-dtor']
-        self.warn_args = {'1': default_warn_args,
+        self.warn_args = {'0': [],
+                          '1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
                           '3': default_warn_args + ['-Wextra']}
 
@@ -303,17 +320,32 @@ class IntelCPPCompiler(IntelCompiler, CPPCompiler):
 
 
 class VisualStudioCPPCompiler(VisualStudioCCompiler, CPPCompiler):
-    def __init__(self, exelist, version, is_cross, exe_wrap, is_64):
+    def __init__(self, exelist, version, is_cross, exe_wrap, target):
         CPPCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
-        VisualStudioCCompiler.__init__(self, exelist, version, is_cross, exe_wrap, is_64)
+        VisualStudioCCompiler.__init__(self, exelist, version, is_cross, exe_wrap, target)
         self.base_options = ['b_pch', 'b_vscrt'] # FIXME add lto, pgo and the like
 
     def get_options(self):
+        cpp_stds = ['none', 'c++11', 'vc++11']
+        if self.id == 'clang-cl':
+            cpp_stds.extend(['c++14', 'vc++14', 'c++17', 'vc++17', 'c++latest'])
+        else:
+            # Visual Studio 2015 and later
+            if version_compare(self.version, '>=19'):
+                cpp_stds.extend(['c++14', 'vc++14', 'c++latest', 'vc++latest'])
+            # Visual Studio 2017 and later
+            if version_compare(self.version, '>=19.11'):
+                cpp_stds.extend(['c++17', 'vc++17'])
+
         opts = CPPCompiler.get_options(self)
         opts.update({'cpp_eh': coredata.UserComboOption('cpp_eh',
                                                         'C++ exception handling type.',
                                                         ['none', 'a', 's', 'sc'],
                                                         'sc'),
+                     'cpp_std': coredata.UserComboOption('cpp_std',
+                                                         'C++ language standard to use',
+                                                         cpp_stds,
+                                                         'none'),
                      'cpp_winlibs': coredata.UserArrayOption('cpp_winlibs',
                                                              'Windows libs to link against.',
                                                              msvc_winlibs)})
@@ -321,9 +353,41 @@ class VisualStudioCPPCompiler(VisualStudioCCompiler, CPPCompiler):
 
     def get_option_compile_args(self, options):
         args = []
-        std = options['cpp_eh']
-        if std.value != 'none':
-            args.append('/EH' + std.value)
+
+        eh = options['cpp_eh']
+        if eh.value != 'none':
+            args.append('/EH' + eh.value)
+
+        vc_version_map = {
+            'none': (True, None),
+            'vc++11': (True, 11),
+            'vc++14': (True, 14),
+            'vc++17': (True, 17),
+            'c++11': (False, 11),
+            'c++14': (False, 14),
+            'c++17': (False, 17)}
+
+        permissive, ver = vc_version_map[options['cpp_std'].value]
+
+        if ver is None:
+            pass
+        elif ver == 11:
+            # Note: there is no explicit flag for supporting C++11; we attempt to do the best we can
+            # which means setting the C++ standard version to C++14, in compilers that support it
+            # (i.e., after VS2015U3)
+            # if one is using anything before that point, one cannot set the standard.
+            if self.id == 'clang-cl' or version_compare(self.version, '>=19.00.24210'):
+                mlog.warning('MSVC does not support C++11; '
+                             'attempting best effort; setting the standard to C++14')
+                args.append('/std:c++14')
+            else:
+                mlog.warning('This version of MSVC does not support cpp_std arguments')
+        else:
+            args.append('/std:c++{}'.format(ver))
+
+        if not permissive and version_compare(self.version, '>=19.11'):
+            args.append('/permissive-')
+
         return args
 
     def get_option_link_args(self, options):
@@ -334,11 +398,15 @@ class VisualStudioCPPCompiler(VisualStudioCCompiler, CPPCompiler):
         # so just use the plain C args.
         return VisualStudioCCompiler.get_compiler_check_args(self)
 
+class ClangClCPPCompiler(VisualStudioCPPCompiler, ClangClCCompiler):
+    def __init__(self, exelist, version, is_cross, exe_wrap, target):
+        VisualStudioCPPCompiler.__init__(self, exelist, version, is_cross, exe_wrap, target)
+        self.id = 'clang-cl'
 
 class ArmCPPCompiler(ArmCompiler, CPPCompiler):
-    def __init__(self, exelist, version, is_cross, exe_wrap=None, **kwargs):
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrap=None, **kwargs):
         CPPCompiler.__init__(self, exelist, version, is_cross, exe_wrap, **kwargs)
-        ArmCompiler.__init__(self)
+        ArmCompiler.__init__(self, compiler_type)
 
     def get_options(self):
         opts = CPPCompiler.get_options(self)
@@ -361,3 +429,31 @@ class ArmCPPCompiler(ArmCompiler, CPPCompiler):
 
     def get_compiler_check_args(self):
         return []
+
+
+class CcrxCPPCompiler(CcrxCompiler, CPPCompiler):
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrap=None, **kwargs):
+        CPPCompiler.__init__(self, exelist, version, is_cross, exe_wrap, **kwargs)
+        CcrxCompiler.__init__(self, compiler_type)
+
+    # Override CCompiler.get_always_args
+    def get_always_args(self):
+        return ['-nologo', '-lang=cpp']
+
+    def get_option_compile_args(self, options):
+        return []
+
+    def get_compile_only_args(self):
+        return []
+
+    def get_output_args(self, target):
+        return ['-output=obj=%s' % target]
+
+    def get_linker_output_args(self, outputname):
+        return ['-output=%s' % outputname]
+
+    def get_option_link_args(self, options):
+        return []
+
+    def get_compiler_check_args(self):
+        return []
