@@ -16,7 +16,10 @@
 functionality such as gobject-introspection, gresources and gtk-doc'''
 
 import os
+import re
+import sys
 import copy
+import shlex
 import subprocess
 
 from .. import build
@@ -24,11 +27,13 @@ from .. import mlog
 from .. import mesonlib
 from .. import compilers
 from .. import interpreter
-from . import GResourceTarget, GResourceHeaderTarget, GirTarget, TypelibTarget, VapiTarget
+from . import GResourceTarget, GResourceHeaderTarget, GResourceObjectTarget, GirTarget, TypelibTarget, VapiTarget
 from . import get_include_args
 from . import ExtensionModule
 from . import ModuleReturnValue
-from ..mesonlib import MesonException, OrderedSet, Popen_safe, extract_as_list
+from ..mesonlib import (
+    MachineChoice, MesonException, OrderedSet, Popen_safe, extract_as_list
+)
 from ..dependencies import Dependency, PkgConfigDependency, InternalDependency
 from ..interpreterbase import noKwargs, permittedKwargs, FeatureNew, FeatureNewKwargs
 
@@ -39,25 +44,28 @@ from ..interpreterbase import noKwargs, permittedKwargs, FeatureNew, FeatureNewK
 # https://bugzilla.gnome.org/show_bug.cgi?id=774368
 gresource_dep_needed_version = '>= 2.51.1'
 
+gresource_ld_binary_needed_version = '>= 2.60'
+
 native_glib_version = None
 girwarning_printed = False
 gdbuswarning_printed = False
 gresource_warning_printed = False
-_gir_has_extra_lib_arg = None
+_gir_has_option = {}
 
-def gir_has_extra_lib_arg(intr_obj):
-    global _gir_has_extra_lib_arg
-    if _gir_has_extra_lib_arg is not None:
-        return _gir_has_extra_lib_arg
+def gir_has_option(intr_obj, option):
+    global _gir_has_option
+    if option in _gir_has_option:
+        return _gir_has_option[option]
 
-    _gir_has_extra_lib_arg = False
+    _gir_has_option[option] = False
     try:
         g_ir_scanner = intr_obj.find_program_impl('g-ir-scanner').get_command()
         opts = Popen_safe(g_ir_scanner + ['--help'], stderr=subprocess.STDOUT)[1]
-        _gir_has_extra_lib_arg = '--extra-library' in opts
+        _gir_has_option[option] = option in opts
     except (MesonException, FileNotFoundError, subprocess.CalledProcessError):
         pass
-    return _gir_has_extra_lib_arg
+
+    return _gir_has_option[option]
 
 class GnomeModule(ExtensionModule):
     gir_dep = None
@@ -153,16 +161,19 @@ class GnomeModule(ExtensionModule):
 
         # Make source dirs relative to build dir now
         source_dirs = [os.path.join(state.build_to_src, state.subdir, d) for d in source_dirs]
-        # Always include current directory, but after paths set by user
-        source_dirs.append(os.path.join(state.build_to_src, state.subdir))
         # Ensure build directories of generated deps are included
         source_dirs += subdirs
+        # Always include current directory, but after paths set by user
+        source_dirs.append(os.path.join(state.build_to_src, state.subdir))
 
         for source_dir in OrderedSet(source_dirs):
             cmd += ['--sourcedir', source_dir]
 
         if 'c_name' in kwargs:
-            cmd += ['--c-name', kwargs.pop('c_name')]
+            c_name = kwargs.pop('c_name')
+            cmd += ['--c-name', c_name]
+        else:
+            c_name = None
         export = kwargs.pop('export', False)
         if not export:
             cmd += ['--internal']
@@ -171,13 +182,19 @@ class GnomeModule(ExtensionModule):
 
         cmd += mesonlib.stringlistify(kwargs.pop('extra_args', []))
 
+        gresource_ld_binary = False
+        if mesonlib.is_linux() and mesonlib.version_compare(glib_version, gresource_ld_binary_needed_version) and not state.environment.is_cross_build():
+            ld_obj = self.interpreter.find_program_impl('ld', required=False)
+            if ld_obj.found():
+                gresource_ld_binary = True
+
         gresource = kwargs.pop('gresource_bundle', False)
-        if gresource:
-            output = args[0] + '.gresource'
-            name = args[0] + '_gresource'
-        else:
-            output = args[0] + '.c'
-            name = args[0] + '_c'
+        if gresource or gresource_ld_binary:
+            g_output = args[0] + '.gresource'
+            g_name = args[0] + '_gresource'
+
+        output = args[0] + '.c'
+        name = args[0] + '_c'
 
         if kwargs.get('install', False) and not gresource:
             raise MesonException('The install kwarg only applies to gresource bundles, see install_header')
@@ -191,18 +208,44 @@ class GnomeModule(ExtensionModule):
         kwargs['input'] = args[1]
         kwargs['output'] = output
         kwargs['depends'] = depends
+        if gresource or gresource_ld_binary:
+            g_kwargs = copy.deepcopy(kwargs)
+            g_kwargs['input'] = args[1]
+            g_kwargs['output'] = g_output
+            g_kwargs['depends'] = depends
         if not mesonlib.version_compare(glib_version, gresource_dep_needed_version):
             # This will eventually go out of sync if dependencies are added
             kwargs['depend_files'] = depend_files
-            kwargs['command'] = cmd
+            if gresource_ld_binary:
+                kwargs['command'] = copy.copy(cmd) + ['--external-data']
+            else:
+                kwargs['command'] = cmd
+            if gresource or gresource_ld_binary:
+                # This will eventually go out of sync if dependencies are added
+                g_kwargs['depend_files'] = depend_files
+                g_kwargs['command'] = cmd
         else:
             depfile = kwargs['output'] + '.d'
-            kwargs['depfile'] = depfile
-            kwargs['command'] = copy.copy(cmd) + ['--dependency-file', '@DEPFILE@']
-        target_c = GResourceTarget(name, state.subdir, state.subproject, kwargs)
+            if gresource_ld_binary:
+                depfile2 = kwargs['output'] + '.2.d'
+                kwargs['depfile'] = depfile2
+                kwargs['command'] = copy.copy(cmd) + ['--external-data', '--dependency-file', '@DEPFILE@']
+            else:
+                kwargs['depfile'] = depfile
+                kwargs['command'] = copy.copy(cmd) + ['--dependency-file', '@DEPFILE@']
+            if gresource or gresource_ld_binary:
+                g_kwargs['depfile'] = depfile
+                g_kwargs['command'] = copy.copy(cmd) + ['--dependency-file', '@DEPFILE@']
 
-        if gresource: # Only one target for .gresource files
-            return ModuleReturnValue(target_c, [target_c])
+        if gresource or gresource_ld_binary:
+            target_g = GResourceTarget(g_name, state.subdir, state.subproject, g_kwargs)
+            if gresource: # Only one target for .gresource files
+                if target_g.get_id() not in self.interpreter.build.targets:
+                    return ModuleReturnValue(target_g, [target_g])
+                else:
+                    return ModuleReturnValue(target_g, [])
+
+        target_c = GResourceTarget(name, state.subdir, state.subproject, kwargs)
 
         h_kwargs = {
             'command': cmd,
@@ -218,9 +261,99 @@ class GnomeModule(ExtensionModule):
             h_kwargs['install_dir'] = kwargs.get('install_dir',
                                                  state.environment.coredata.get_builtin_option('includedir'))
         target_h = GResourceHeaderTarget(args[0] + '_h', state.subdir, state.subproject, h_kwargs)
-        rv = [target_c, target_h]
+
+        if gresource_ld_binary:
+            return self._create_gresource_ld_binary_targets(args, state, ifile, ld_obj, c_name, target_g, g_output, target_c, target_h)
+        else:
+            rv = [target_c, target_h]
+
         return ModuleReturnValue(rv, rv)
 
+    def _create_gresource_ld_binary_targets(self, args, state, ifile, ld_obj, c_name, target_g, g_output, target_c, target_h):
+        if c_name is None:
+            # Create proper c identifier from filename in the way glib-compile-resources does
+            c_name = os.path.basename(ifile).partition('.')[0]
+            c_name = c_name.replace('-', '_')
+            c_name = re.sub(r'^([^(_a-zA-Z)])+', '', c_name)
+            c_name = re.sub(r'([^(_a-zA-Z0-9)])', '', c_name)
+
+        c_name_no_underscores = re.sub(r'^_+', '', c_name)
+
+        ld = ld_obj.get_command()
+        objcopy_object = self.interpreter.find_program_impl('objcopy', required=False)
+        if objcopy_object.found():
+            objcopy = objcopy_object.get_command()
+        else:
+            objcopy = None
+
+        o_kwargs = {
+            'command': [ld, '-r', '-b', 'binary', '@INPUT@', '-o', '@OUTPUT@'],
+            'input': target_g,
+            'output': args[0] + '1.o'
+        }
+
+        target_o = GResourceObjectTarget(args[0] + '1_o', state.subdir, state.subproject, o_kwargs)
+
+        builddir = os.path.join(state.environment.get_build_dir(), state.subdir)
+        linkerscript_name = args[0] + '_map.ld'
+        linkerscript_path = os.path.join(builddir, linkerscript_name)
+        linkerscript_file = open(linkerscript_path, 'w')
+
+        # Create symbol name the way bfd does
+        binary_name = os.path.join(state.subdir, g_output)
+        encoding = sys.getfilesystemencoding()
+        symbol_name = re.sub(rb'([^(_a-zA-Z0-9)])', b'_', binary_name.encode(encoding)).decode(encoding)
+
+        linkerscript_string = '''SECTIONS
+{{
+  .gresource.{} : ALIGN(8)
+  {{
+    {}_resource_data = _binary_{}_start;
+  }}
+  .data :
+  {{
+    *(.data)
+  }}
+}}'''.format(c_name_no_underscores, c_name, symbol_name)
+
+        linkerscript_file.write(linkerscript_string)
+
+        o2_kwargs = {
+            'command': [ld, '-r', '-T', os.path.join(state.subdir, linkerscript_name), '@INPUT@', '-o', '@OUTPUT@'],
+            'input': target_o,
+            'output': args[0] + '2.o',
+        }
+        target_o2 = GResourceObjectTarget(args[0] + '2_o', state.subdir, state.subproject, o2_kwargs)
+
+        if objcopy is not None:
+            objcopy_cmd = [objcopy, '--set-section-flags', '.gresource.' + c_name + '=readonly,alloc,load,data']
+            objcopy_cmd += ['-N', '_binary_' + symbol_name + '_start']
+            objcopy_cmd += ['-N', '_binary_' + symbol_name + '_end']
+            objcopy_cmd += ['-N', '_binary_' + symbol_name + '_size']
+            objcopy_cmd += ['@INPUT@', '@OUTPUT@']
+
+            o3_kwargs = {
+                'command': objcopy_cmd,
+                'input': target_o2,
+                'output': args[0] + '3.o'
+            }
+
+            target_o3 = GResourceObjectTarget(args[0] + '3_o', state.subdir, state.subproject, o3_kwargs)
+
+            rv1 = [target_c, target_h, target_o3]
+            if target_g.get_id() not in self.interpreter.build.targets:
+                rv2 = rv1 + [target_g, target_o, target_o2]
+            else:
+                rv2 = rv1 + [target_o, target_o2]
+        else:
+            rv1 = [target_c, target_h, target_o2]
+            if target_g.get_id() not in self.interpreter.build.targets:
+                rv2 = rv1 + [target_g, target_o]
+            else:
+                rv2 = rv1 + [target_o]
+
+        return ModuleReturnValue(rv1, rv2)
+
     def _get_gresource_dependencies(self, state, input_file, source_dirs, dependencies):
 
         cmd = ['glib-compile-resources',
@@ -307,7 +440,7 @@ class GnomeModule(ExtensionModule):
             if include_rpath:
                 link_command.append('-Wl,-rpath,' + libdir)
             depends.append(lib)
-        if gir_has_extra_lib_arg(self.interpreter) and use_gir_args:
+        if gir_has_option(self.interpreter, '--extra-library') and use_gir_args:
             link_command.append('--extra-library=' + lib.name)
         else:
             link_command.append('-l' + lib.name)
@@ -326,19 +459,20 @@ class GnomeModule(ExtensionModule):
 
         for dep in deps:
             if isinstance(dep, InternalDependency):
-                cflags.update(dep.compile_args)
+                cflags.update(dep.get_compile_args())
                 cflags.update(get_include_args(dep.include_directories))
                 for lib in dep.libraries:
                     if hasattr(lib, 'held_object'):
                         lib = lib.held_object
-                    internal_ldflags.update(self._get_link_args(state, lib, depends, include_rpath))
-                    libdepflags = self._get_dependencies_flags(lib.get_external_deps(), state, depends, include_rpath,
-                                                               use_gir_args, True)
-                    cflags.update(libdepflags[0])
-                    internal_ldflags.update(libdepflags[1])
-                    external_ldflags.update(libdepflags[2])
-                    external_ldflags_nodedup += libdepflags[3]
-                    gi_includes.update(libdepflags[4])
+                    if isinstance(lib, build.SharedLibrary):
+                        internal_ldflags.update(self._get_link_args(state, lib, depends, include_rpath))
+                        libdepflags = self._get_dependencies_flags(lib.get_external_deps(), state, depends, include_rpath,
+                                                                   use_gir_args, True)
+                        cflags.update(libdepflags[0])
+                        internal_ldflags.update(libdepflags[1])
+                        external_ldflags.update(libdepflags[2])
+                        external_ldflags_nodedup += libdepflags[3]
+                        gi_includes.update(libdepflags[4])
                 extdepflags = self._get_dependencies_flags(dep.ext_deps, state, depends, include_rpath,
                                                            use_gir_args, True)
                 cflags.update(extdepflags[0])
@@ -390,7 +524,7 @@ class GnomeModule(ExtensionModule):
                 mlog.log('dependency {!r} not handled to build gir files'.format(dep))
                 continue
 
-        if gir_has_extra_lib_arg(self.interpreter) and use_gir_args:
+        if gir_has_option(self.interpreter, '--extra-library') and use_gir_args:
             def fix_ldflags(ldflags):
                 fixed_ldflags = OrderedSet()
                 for ldflag in ldflags:
@@ -406,11 +540,21 @@ class GnomeModule(ExtensionModule):
         else:
             return cflags, internal_ldflags, external_ldflags, external_ldflags_nodedup, gi_includes
 
-    def _unwrap_gir_target(self, girtarget):
+    def _unwrap_gir_target(self, girtarget, state):
         while hasattr(girtarget, 'held_object'):
             girtarget = girtarget.held_object
-        if not isinstance(girtarget, (build.Executable, build.SharedLibrary)):
-            raise MesonException('Gir target must be an executable or shared library')
+
+        if not isinstance(girtarget, (build.Executable, build.SharedLibrary,
+                                      build.StaticLibrary)):
+            raise MesonException('Gir target must be an executable or library')
+
+        STATIC_BUILD_REQUIRED_VERSION = ">=1.58.1"
+        if isinstance(girtarget, (build.StaticLibrary)) and \
+           not mesonlib.version_compare(
+               self._get_gir_dep(state)[0].get_version(),
+               STATIC_BUILD_REQUIRED_VERSION):
+            raise MesonException('Static libraries can only be introspected with GObject-Introspection ' + STATIC_BUILD_REQUIRED_VERSION)
+
         return girtarget
 
     def _get_gir_dep(self, state):
@@ -518,7 +662,8 @@ class GnomeModule(ExtensionModule):
         ret = []
 
         for lang in langs:
-            for link_arg in state.environment.coredata.get_external_link_args(lang):
+            link_args = state.environment.coredata.get_external_link_args(MachineChoice.HOST, lang)
+            for link_arg in link_args:
                 if link_arg.startswith('-L'):
                     ret.append(link_arg)
 
@@ -530,8 +675,13 @@ class GnomeModule(ExtensionModule):
         for girtarget in girtargets:
             if isinstance(girtarget, build.Executable):
                 ret += ['--program', girtarget]
-            elif isinstance(girtarget, build.SharedLibrary):
-                libname = girtarget.get_basename()
+            else:
+                # Because of https://gitlab.gnome.org/GNOME/gobject-introspection/merge_requests/72
+                # we can't use the full path until this is merged.
+                if isinstance(girtarget, build.SharedLibrary):
+                    libname = girtarget.get_basename()
+                else:
+                    libname = os.path.join("@PRIVATE_OUTDIR_ABS_%s@" % girtarget.get_id(), girtarget.get_filename())
                 # Needed for the following binutils bug:
                 # https://github.com/mesonbuild/meson/issues/1911
                 # However, g-ir-scanner does not understand -Wl,-rpath
@@ -584,9 +734,15 @@ class GnomeModule(ExtensionModule):
             if 'b_sanitize' in compiler.base_options:
                 sanitize = state.environment.coredata.base_options['b_sanitize'].value
                 cflags += compilers.sanitizer_compile_args(sanitize)
-                if 'address' in sanitize.split(','):
-                    internal_ldflags += ['-lasan']  # This must be first in ldflags
-                # FIXME: Linking directly to libasan is not recommended but g-ir-scanner
+                sanitize = sanitize.split(',')
+                # These must be first in ldflags
+                if 'address' in sanitize:
+                    internal_ldflags += ['-lasan']
+                if 'thread' in sanitize:
+                    internal_ldflags += ['-ltsan']
+                if 'undefined' in sanitize:
+                    internal_ldflags += ['-lubsan']
+                # FIXME: Linking directly to lib*san is not recommended but g-ir-scanner
                 # does not understand -f LDFLAGS. https://bugzilla.gnome.org/show_bug.cgi?id=783892
                 # ldflags += compilers.sanitizer_link_args(sanitize)
 
@@ -691,7 +847,7 @@ class GnomeModule(ExtensionModule):
     def _get_external_args_for_langs(self, state, langs):
         ret = []
         for lang in langs:
-            ret += state.environment.coredata.get_external_args(lang)
+            ret += state.environment.coredata.get_external_args(MachineChoice.HOST, lang)
         return ret
 
     @staticmethod
@@ -722,7 +878,7 @@ class GnomeModule(ExtensionModule):
         giscanner = self.interpreter.find_program_impl('g-ir-scanner')
         gicompiler = self.interpreter.find_program_impl('g-ir-compiler')
 
-        girtargets = [self._unwrap_gir_target(arg) for arg in args]
+        girtargets = [self._unwrap_gir_target(arg, state) for arg in args]
 
         if len(girtargets) > 1 and any([isinstance(el, build.Executable) for el in girtargets]):
             raise MesonException('generate_gir only accepts a single argument when one of the arguments is an executable')
@@ -780,6 +936,10 @@ class GnomeModule(ExtensionModule):
         scan_command += self._scan_langs(state, [lc[0] for lc in langs_compilers])
         scan_command += list(external_ldflags)
 
+        if gir_has_option(self.interpreter, '--sources-top-dirs'):
+            scan_command += ['--sources-top-dirs', os.path.join(state.environment.get_source_dir(), self.interpreter.subproject_dir, state.subproject)]
+            scan_command += ['--sources-top-dirs', os.path.join(state.environment.get_build_dir(), self.interpreter.subproject_dir, state.subproject)]
+
         scan_target = self._make_gir_target(state, girfile, scan_command, depends, kwargs)
 
         typelib_output = '%s-%s.typelib' % (ns, nsversion)
@@ -883,12 +1043,14 @@ This will become a hard error in the future.''')
         return ModuleReturnValue(None, rv)
 
     @FeatureNewKwargs('gnome.gtkdoc', '0.48.0', ['c_args'])
+    @FeatureNewKwargs('gnome.gtkdoc', '0.48.0', ['module_version'])
     @FeatureNewKwargs('gnome.gtkdoc', '0.37.0', ['namespace', 'mode'])
     @permittedKwargs({'main_xml', 'main_sgml', 'src_dir', 'dependencies', 'install',
                       'install_dir', 'scan_args', 'scanobjs_args', 'gobject_typesfile',
                       'fixxref_args', 'html_args', 'html_assets', 'content_files',
                       'mkdb_args', 'ignore_headers', 'include_directories',
-                      'namespace', 'mode', 'expand_content_files'})
+                      'namespace', 'mode', 'expand_content_files', 'module_version',
+                      'c_args'})
     def gtkdoc(self, state, args, kwargs):
         if len(args) != 1:
             raise MesonException('Gtkdoc must have one positional argument.')
@@ -903,11 +1065,14 @@ This will become a hard error in the future.''')
         main_xml = kwargs.get('main_xml', '')
         if not isinstance(main_xml, str):
             raise MesonException('Main xml keyword argument must be a string.')
+        moduleversion = kwargs.get('module_version', '')
+        if not isinstance(moduleversion, str):
+            raise MesonException('Module version keyword argument must be a string.')
         if main_xml != '':
             if main_file != '':
                 raise MesonException('You can only specify main_xml or main_sgml, not both.')
             main_file = main_xml
-        targetname = modulename + '-doc'
+        targetname = modulename + ('-' + moduleversion if moduleversion else '') + '-doc'
         command = state.environment.get_build_command()
 
         namespace = kwargs.get('namespace', '')
@@ -938,6 +1103,7 @@ This will become a hard error in the future.''')
                 '--headerdirs=' + '@@'.join(header_dirs),
                 '--mainfile=' + main_file,
                 '--modulename=' + modulename,
+                '--moduleversion=' + moduleversion,
                 '--mode=' + mode]
         if namespace:
             args.append('--namespace=' + namespace)
@@ -1006,20 +1172,24 @@ This will become a hard error in the future.''')
         ldflags.update(internal_ldflags)
         ldflags.update(external_ldflags)
 
+        cflags.update(state.environment.coredata.get_external_args(MachineChoice.HOST, 'c'))
+        ldflags.update(state.environment.coredata.get_external_link_args(MachineChoice.HOST, 'c'))
         if state.environment.is_cross_build():
             compiler = state.environment.coredata.cross_compilers.get('c')
         else:
-            cflags.update(state.environment.coredata.get_external_args('c'))
-            ldflags.update(state.environment.coredata.get_external_link_args('c'))
             compiler = state.environment.coredata.compilers.get('c')
 
+        compiler_flags = self._get_langs_compilers_flags(state, [('c', compiler)])
+        cflags.update(compiler_flags[0])
+        ldflags.update(compiler_flags[1])
+        ldflags.update(compiler_flags[2])
         if compiler:
-            args += ['--cc=%s' % ' '.join(compiler.get_exelist())]
-            args += ['--ld=%s' % ' '.join(compiler.get_linker_exelist())]
+            args += ['--cc=%s' % ' '.join([shlex.quote(x) for x in compiler.get_exelist()])]
+            args += ['--ld=%s' % ' '.join([shlex.quote(x) for x in compiler.get_linker_exelist()])]
         if cflags:
-            args += ['--cflags=%s' % ' '.join(cflags)]
+            args += ['--cflags=%s' % ' '.join([shlex.quote(x) for x in cflags])]
         if ldflags:
-            args += ['--ldflags=%s' % ' '.join(ldflags)]
+            args += ['--ldflags=%s' % ' '.join([shlex.quote(x) for x in ldflags])]
 
         return args
 
